#ifndef GAME_H_
#define GAME_H_
/* 
notes on developing the implementation of this interface for a specific platform:
- the implementation should call srand with an appopriate seed (different each time, needn't be perfectly random)
- the bottom of the screen is always y = 0. this means you'll probably have to flip the y coordinate of the mouse
*/

#ifdef DEBUG
#undef DEBUG
#define DEBUG 1
#endif

#ifdef PROFILE
#undef PROFILE
#define PROFILE 1
#endif

#if !DEBUG
#define NDEBUG
#endif

#include <limits.h>
#include <stddef.h>

#ifdef _WIN32
#include "lib/glcorearb.h"
#endif

#ifdef __GNUC__
#define no_warn_start _Pragma("GCC diagnostic push") \
	_Pragma("GCC diagnostic ignored \"-Wpedantic\"") \
	_Pragma("GCC diagnostic ignored \"-Wsign-conversion\"") \
	_Pragma("GCC diagnostic ignored \"-Wsign-compare\"") \
	_Pragma("GCC diagnostic ignored \"-Wconversion\"") \
	_Pragma("GCC diagnostic ignored \"-Wimplicit-fallthrough\"") \
	_Pragma("GCC diagnostic ignored \"-Wunused-function\"")

#define no_warn_end _Pragma("GCC diagnostic pop")
#else
#define no_warn_start
#define no_warn_end
#endif

#ifndef __cplusplus
#if defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L
typedef _Bool bool;
#else
typedef unsigned char bool;
#endif
#define false 0
#define true 1
#endif

typedef unsigned long ulong;
typedef unsigned int  uint;

typedef unsigned char  u8;
typedef   signed char  i8;
typedef unsigned short u16;
typedef          short i16;
#if UINT_MAX < 4294967295
typedef unsigned long  u32;
typedef          long  i32;
#define U32_FMT "%lu"
#else
typedef unsigned int   u32;
typedef          int   i32;
#define U32_FMT "%u"
#endif

#define U8_MAX 255
#define U16_MAX 65535
#define U32_MAX 4294967295

#define I8_MAX 127
#define I16_MAX 32767
#define I32_MAX 2147483647

#define GAME_MEMORY_REQUIRED (16L<<20)

// upper 12 bits of the port used to communicate. lower 4 bits may vary. 
#define PORT_UPPER_PART 0x133

#define arr_count(arr) (sizeof (arr) / sizeof *(arr))

enum {
	KEY_UNKNOWN,
	KEY_UP,
	KEY_LEFT,
	KEY_RIGHT,
	KEY_DOWN,
	KEY_PAGEDOWN,
	KEY_PAGEUP,
	KEY_SPACE,
	KEY_BACKSPACE,
	KEY_ESCAPE,
	KEY_A,
	KEY_B,
	KEY_C,
	KEY_D,
	KEY_E,
	KEY_F,
	KEY_G,
	KEY_H,
	KEY_I,
	KEY_J,
	KEY_K,
	KEY_L,
	KEY_M,
	KEY_N,
	KEY_O,
	KEY_P,
	KEY_Q,
	KEY_R,
	KEY_S,
	KEY_T,
	KEY_U,
	KEY_V,
	KEY_W,
	KEY_X,
	KEY_Y,
	KEY_Z,
	KEY_0,
	KEY_1,
	KEY_2,
	KEY_3,
	KEY_4,
	KEY_5,
	KEY_6,
	KEY_7,
	KEY_8,
	KEY_9,
	KEY_F1,
	KEY_F2,
	KEY_F3,
	KEY_F4,
	KEY_F5,
	KEY_F6,
	KEY_F7,
	KEY_F8,
	KEY_F9,
	KEY_F10,
	KEY_F11,
	KEY_F12,
	KEY_LCTRL,
	KEY_RCTRL,
	KEY_LSHIFT,
	KEY_RSHIFT,
	KEY_ENTER,
	NKEYS
};

typedef enum {
	MBUTTON_OTHER,
	MBUTTON_LEFT,
	MBUTTON_MIDDLE,
	MBUTTON_RIGHT
} MouseButton;

enum {
	MBUTTON_MASK_LEFT = 1L<<MBUTTON_LEFT,
	MBUTTON_MASK_MIDDLE = 1L<<MBUTTON_MIDDLE,
	MBUTTON_MASK_RIGHT = 1L<<MBUTTON_RIGHT
};

typedef struct {
	int x, y; // bottom is y = 0 
	u8 button;
} MouseEvent;

typedef struct {
	u16 key;
} KeyEvent;

typedef struct {
	char text[16];
} TextEvent;

typedef struct {
	int x, y; // bottom is y = 0 
	u8 mask; // mask of mouse buttons down 
} MouseMoveEvent;

typedef struct {
	int dy; // +1 = up, -1 = down
} ScrollEvent;

enum {
	EVENT_KEY_UP,
	EVENT_KEY_DOWN,
	EVENT_MOUSE_UP,
	EVENT_MOUSE_DOWN,
	EVENT_MOUSE_MOVE,
	EVENT_SCROLL,
	EVENT_TEXT
};

typedef struct {
	u8 kind; // see enum above 
	union {
		KeyEvent key;
		MouseEvent mouse;
		MouseMoveEvent mmove;
		TextEvent text;
		ScrollEvent scroll;
	} u;
} Event;

#define MAX_EVENTS_PER_FRAME 2048
typedef struct {
	bool focus; // does the window have focus? 
	int mouse_x, mouse_y; // bottom is y = 0 
	u8 mouse_mask; // mask of MBUTTON_MASK_* 
	bool closed; // was the window closed by the user? 
	u16 nevents;
	bool keys_down[NKEYS];
	Event events[MAX_EVENTS_PER_FRAME];
} Input;

typedef struct Server Server;
typedef struct Client Client;

typedef struct {
	u8 a,b,c,d;
} IPv4;

typedef struct {
	char unused;
} ClientInfo;

typedef struct {
	IPv4 ip;
	u16 port;
	u16 password;
} ServerInfo;

typedef struct {
	void (*(*get_gl_proc)(char const *))(void);
	/* 
		lists all *files* (not including directories) in the given directory in a NULL-terminated array. returns NULL on failure.
		the return value must be allocated in such away that it and all of its elements can be freed with free
	*/
	char **(*list_files)(char const *dir);
	/* 
		returns:
		1 - if the directory was created successfully
		0 - if there is already a *directory* (NOT a file) at that path.
		-1 - if an error occured
	*/
	int (*mkdir)(char const *dir);
	// communication functions 
	Server *(*server_create)(ServerInfo *info); // create a server and start waiting for clients, also fill out server info 
	bool (*server_check_for_client)(Server *server, ClientInfo *info); // check if there's a client; return true and fill out info if there is 
	/* 
		The send/recv functions should all return -1 if a serious error occurs (e.g. connection lost),
		and 0 if something non-serious happens (try again later, etc.)
	*/
	long (*server_send)(Server *server, void const *data, size_t n); // returns number of bytes sent 
	long (*server_recv)(Server *server, void *data, size_t n); // returns number of bytes received 
	void (*server_disconnect_from_client)(Server *server);
	void (*server_destroy)(Server *server); // stop and free the server 
	Client *(*client_create)(ServerInfo *server_info); // create a client and start connecting to the server 
	int (*client_is_connected)(Client *client); // returns 0 if not connected, 1 if connected, -1 if an error occured 
	long (*client_send)(Client *client, void const *data, size_t n); // returns number of bytes sent 
	long (*client_recv)(Client *client, void *data, size_t n); // receive data from the server. returns number of bytes received 
	void (*client_destroy)(Client *client); // destroy the client 
	char *(*comm_geterr)(void); // get error from server/client functions 
} PlatformProcs;

typedef struct {
	int width, height;
	ulong memory_size;
	void *memory;
	PlatformProcs procs;
	Input input;
	bool close; // should the window be closed? (defaults to input.closed) 
	bool fullscreen; // should the window be fullscreen? 
	double dt;
} Frame;
#endif // GAME_H_ 
