// generic "invalid hex" 
static HexPos const HEX_POS_INVALID = {I16_MAX, I16_MAX};

static void hex_print(HexPos h) {
	printf("h(%u, %u)\n", h.x, h.y);
}

// is this hex position actually in the grid? 
static bool hex_in_bounds(Game *game, HexPos h) {
	return h.x < game->width && h.y < game->height;
}

static v2 hex_center_pos2_(int col, int row, float hex_radius, v2 grid_base) {
	float r = hex_radius;
	float h = r * HALF_SQRT3f;
	float x, y;

	y = h + 2.0f*h*(float)row; // initial offset = h, add 2 * h per row we go upwards 
	if (col % 2 == 1)
		y += h; // bump up every odd column 
	x = r + 1.5f*r*(float)col; // initial offset = r, add 1.5 * r per column we go to the right 
	return v2_add(V2(x,y), grid_base);
}

static v2 hex_center_pos_(HexPos pos, float hex_radius, v2 grid_base) {
	return hex_center_pos2_(pos.x, pos.y, hex_radius, grid_base);
}

// position of center of hexagon at (col, row) 
static v2 hex_center_pos2(Game *game, int col, int row) {
	return hex_center_pos2_(col, row, game->hex_radius, game->grid_base);
}

static v2 hex_center_pos(Game *game, HexPos pos) {
	return hex_center_pos2(game, pos.x, pos.y);
}

static HexPos hex_at_pos_(v2 pos, float hex_radius, v2 grid_base) {
	float x = pos.x, y = pos.y;
	float r = hex_radius;
	float h = r * HALF_SQRT3f;
	int p1x, p1y, p2x, p2y, p3x, p3y;
	v2 v, pos1, pos2, pos3;
	float d1, d2, d3;

	if (hex_radius == 0) {
		return HEX_POS_INVALID;
	}

	v = V2(x,y);

	// remove base offset 
	x -= grid_base.x;
	y -= grid_base.y;

	// calculate a guess 
	p1x = (int)floorf((x-r*0.5f) / (1.5f*r));
	if (p1x & 1)
		y -= h;
	p1y = (int)floorf(y / (2.0f*h));
	// this guess might not be right, it could also be the one up and to the right or down and to the right 

	// other candidates 
	// sometimes up right, sometimes down right 
	p2x = p1x+1;
	p2y = p1y;
	if (p1x & 1) {
		// up right 
		p3x = p1x+1;
		p3y = p1y+1;
	} else {
		// down right 
		p3x = p1x+1;
		p3y = p1y-1;
	}

	// positions of all candidates 
	pos1 = hex_center_pos2_(p1x, p1y, hex_radius, grid_base);
	pos2 = hex_center_pos2_(p2x, p2y, hex_radius, grid_base);
	pos3 = hex_center_pos2_(p3x, p3y, hex_radius, grid_base);
	
	// distances (squared) to all candidates 
	d1 = v2_dist_squared(v, pos1);
	d2 = v2_dist_squared(v, pos2);
	d3 = v2_dist_squared(v, pos3);

	// return candidate closest to point 
	if (d1 < d2 && d1 < d3) {
		return hex_pos((u16)p1x, (u16)p1y);
	} else if (d2 < d1 && d2 < d3) {
		return hex_pos((u16)p2x, (u16)p2y);
	} else {
		return hex_pos((u16)p3x, (u16)p3y);
	}
}

static v2 hex_grid_size(float hex_radius, int grid_width, int grid_height) {
	float r = hex_radius;
	float h = SQRT3f * r;
	float W = (float)grid_width, H = (float)grid_height;
	return V2(0.5f * r * (3*W+1), (H+0.5f) * h);
}

static Rect game_grid_bounds(Game *game) {
	v2 pos = game->grid_base;
	v2 size = hex_grid_size(game->hex_radius, game->width, game->height);
	return rect(pos, size);
}

static HexPos hex_at_pos(Game *game, v2 pos) {
	return hex_at_pos_(pos, game->hex_radius, game->grid_base);
}

static bool hex_eq(HexPos a, HexPos b) {
	return a.x == b.x && a.y == b.y;
}

static u32 hex_idx(Game *game, HexPos p) {
	return (u32)p.y * game->width + (u32)p.x;
}

static u8 hex_tile_kind(Game *game, HexPos pos) {
	assert(hex_in_bounds(game, pos));
	return game->tiles[hex_idx(game, pos)];
}

// can this hex pos have a pion at it?
static bool hex_can_have_pion(Game *game, HexPos pos) {
	if (!hex_in_bounds(game, pos))
		return false;
	return hex_tile_kind(game, pos) != TL_NONE;
}

static bool hexes_are_adjacent(HexPos a, HexPos b) {
	// is a adjacent to b?
	if (b.x == a.x) {
		return b.y == a.y+1 || b.y == a.y-1;
	} else if (b.y == a.y) {
		return b.x == a.x+1 || b.x == a.x-1;
	} else if (a.x & 1) {
		// a is an up hex 
		return b.y == a.y+1 && (b.x == a.x-1 || b.x == a.x+1);
	} else {
		// a is a down hex 
		return b.y == a.y-1 && (b.x == a.x-1 || b.x == a.x+1);
	}
}

/*
find the right hex radius 
grid_min_coords = lower left corner, grid_max_coords = upper right corner of area where grid can go
*/
static void hex_setup_game_grid(Game *game, v2 grid_min_coords, v2 grid_max_coords) {
	float min_x = grid_min_coords.x, min_y = grid_min_coords.y;
	float max_x = grid_max_coords.x, max_y = grid_max_coords.y;
	float X = max_x - min_x, Y = max_y - min_y;
	float W = (float)game->width, H = (float)game->height;
	float hex_radius1 = (2*X)/(3*W+1);
	float hex_radius2 = (SQRT3f * (2/3.0f)) * (Y / (2*H+1));

	// find radius that fits both x and y 
	game->hex_radius = minf(hex_radius1, hex_radius2);
	{
		v2 grid_size = hex_grid_size(game->hex_radius, game->width, game->height);
		float center_x = lerpf(0.5f, min_x, max_x);
		float center_y = lerpf(0.5f, min_y, max_y);
		// center grid within bounds 
		float base_x = center_x - 0.5f * grid_size.x;
		float base_y = center_y - 0.5f * grid_size.y;

		game->grid_base = V2(base_x, base_y);
	}

}

