#include "game.c"

#ifdef _WIN32
#include <winsock2.h>
#include <iphlpapi.h>
#pragma comment(lib, "iphlpapi.lib")
#pragma comment(lib, "ws2_32.lib")
#pragma comment(lib, "opengl32.lib")
#include <windows.h>
#include "SDL.h"
#else
#include <SDL2/SDL.h>
#endif
#include <GL/gl.h>

#ifdef __unix__
#include "sockets-posix.c"
#include "filesystem-posix.c"
#elif defined _WIN32
#include "sockets-win.c"
#include "filesystem-win.c"
#else
#error "Unrecognized operating system."
#endif

#include "procs.c"

static SDL_Window *sdl_window = NULL;

static u16 keycode_to_key(SDL_Keycode sym) {
	switch (sym) {
	case SDLK_LEFT: return KEY_LEFT;
	case SDLK_RIGHT: return KEY_RIGHT;
	case SDLK_UP: return KEY_UP;
	case SDLK_DOWN: return KEY_DOWN;
	case SDLK_SPACE: return KEY_SPACE;
	case SDLK_LCTRL: return KEY_LCTRL;
	case SDLK_RCTRL: return KEY_RCTRL;
	case SDLK_LSHIFT: return KEY_LSHIFT;
	case SDLK_RSHIFT: return KEY_RSHIFT;
	case SDLK_RETURN: return KEY_ENTER;
	case SDLK_BACKSPACE: return KEY_BACKSPACE;
	case SDLK_ESCAPE: return KEY_ESCAPE;
	case SDLK_PAGEDOWN: return KEY_PAGEDOWN;
	case SDLK_PAGEUP: return KEY_PAGEUP;
	default:
		if (sym >= SDLK_a && sym <= SDLK_z) {
			return (u16)(sym - SDLK_a + KEY_A);
		} else if (sym >= SDLK_0 && sym <= SDLK_9) {
			return (u16)(sym - SDLK_0 + KEY_0);
		} else if (sym >= SDLK_F1 && sym <= SDLK_F12) {
			return (u16)(sym - SDLK_F1 + KEY_F1);
		}
	}
	return KEY_UNKNOWN;
}

static SDL_Scancode key_to_scancode(u16 key) {
	switch (key) {
	case KEY_LEFT: return SDL_SCANCODE_LEFT;
	case KEY_RIGHT: return SDL_SCANCODE_RIGHT;
	case KEY_UP: return SDL_SCANCODE_UP;
	case KEY_DOWN: return SDL_SCANCODE_DOWN;
	case KEY_SPACE: return SDL_SCANCODE_SPACE;
	case KEY_LCTRL: return SDL_SCANCODE_LCTRL;
	case KEY_RCTRL: return SDL_SCANCODE_RCTRL;
	case KEY_LSHIFT: return SDL_SCANCODE_LSHIFT;
	case KEY_RSHIFT: return SDL_SCANCODE_RSHIFT;
	case KEY_ENTER: return SDL_SCANCODE_RETURN;
	case KEY_BACKSPACE: return SDL_SCANCODE_BACKSPACE;
	case KEY_ESCAPE: return SDL_SCANCODE_ESCAPE;
	case KEY_PAGEDOWN: return SDL_SCANCODE_PAGEDOWN;
	case KEY_PAGEUP: return SDL_SCANCODE_PAGEUP;
	case KEY_0: return SDL_SCANCODE_0; // SDL_SCANCODE_0 != SDL_SCANCODE_1 - 1 
	default:
		if (key >= KEY_A && key <= KEY_Z) {
			return SDL_SCANCODE_A + (key - KEY_A);
		} else if (key >= KEY_1 && key <= KEY_9) {
			return SDL_SCANCODE_1 + (key - KEY_1);
		} else if (key >= KEY_F1 && key <= KEY_F12) {
			return SDL_SCANCODE_F1 + (key - KEY_F1);
		}
		break;
	}
	return 0;
}

static void die(char const *fmt, ...) {
	char message[1024] = {0};
	va_list args;
	va_start(args, fmt);
	vsprintf(message, fmt, args);
	va_end(args);
	if (sdl_window) SDL_DestroyWindow(sdl_window);
	SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, "Error", message, NULL);
	exit(-1);
}

static u8 convert_mouse_mask(uint mask) {
	u8 mouse_mask = 0;
	if (mask & SDL_BUTTON_LMASK)
		mouse_mask |= MBUTTON_MASK_LEFT;
	if (mask & SDL_BUTTON_MMASK)
		mouse_mask |= MBUTTON_MASK_MIDDLE;
	if (mask & SDL_BUTTON_RMASK)
		mouse_mask |= MBUTTON_MASK_RIGHT;
	return mouse_mask;
}

static void add_event(Input *input, Event *e) {
	if (input->nevents < MAX_EVENTS_PER_FRAME) {
		input->events[input->nevents++] = *e;
	}
}

#ifdef _WIN32
INT WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
    PSTR lpCmdLine, INT nCmdShow)
#else
int main(void)
#endif
{
	SDL_Window *window = NULL;
	SDL_GLContext glctx = NULL;
	Frame frame = {0};
	bool focus = true;
	bool fullscreen = frame.fullscreen = true;
	Uint32 last_frame_ticks = 0;
	Input *input = &frame.input;
	int dfl_windowed_width = 1280, dfl_windowed_height = 720;

	SDL_SetHint(SDL_HINT_NO_SIGNAL_HANDLERS, "1"); // if this program is sent a SIGTERM/SIGINT, don't turn it into a quit event 
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		die("Couldn't initialize SDL: %s.", SDL_GetError());
	}
#if _WIN32
	{
		WSADATA wsa_data;
		if (WSAStartup(MAKEWORD(2,2), &wsa_data) != 0) {
			printf("Error: WSAStartup failed.\n");
			return 1;
		}
	}
#endif
	sdl_window = window = SDL_CreateWindow("PIONS", 0, 0, dfl_windowed_width, dfl_windowed_height,
		SDL_WINDOW_OPENGL|SDL_WINDOW_SHOWN|SDL_WINDOW_FULLSCREEN_DESKTOP);

	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 2);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 0);
	glctx = SDL_GL_CreateContext(window);
	if (!glctx) {
		die("%s", SDL_GetError());
	}
	
	{
		ulong memory_size = frame.memory_size = GAME_MEMORY_REQUIRED;
		void *memory = frame.memory = calloc(1, memory_size);
		if (!memory) {
			die("You don't have enough available memory to run this game. Try closing some stuff.");
		}
	}
	
	last_frame_ticks = SDL_GetTicks();

	fill_out_procs(&frame.procs, (void (*(*)(char const *))(void))SDL_GL_GetProcAddress);
	{ // not cryptographically secure or anything, but okay 
		unsigned a = (unsigned)last_frame_ticks;
		unsigned b = (unsigned)time(NULL);
		srand(a*b);
	}

	SDL_GL_SetSwapInterval(1); // vsync (seems to not be on by default) 
    SDL_StartTextInput();

	while (1) {
		SDL_Event event;
		glClearColor(0,0,0,1);
		glClear(GL_COLOR_BUFFER_BIT);

		memset(input, 0, sizeof *input);
		SDL_GetWindowSize(window, &frame.width, &frame.height);
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				input->closed = true;
				break;
			case SDL_WINDOWEVENT:
				switch (event.window.event) {
				case SDL_WINDOWEVENT_FOCUS_GAINED:
					focus = true;
					break;
				case SDL_WINDOWEVENT_FOCUS_LOST:
					focus = false;
					break;
				}
				break;
			case SDL_KEYDOWN:
			case SDL_KEYUP: {
				Event e = {0};
				u16 key = keycode_to_key(event.key.keysym.sym);
				KeyEvent *k = &e.u.key;
				e.kind = event.type == SDL_KEYDOWN ? EVENT_KEY_DOWN : EVENT_KEY_UP;
				k->key = key;
				add_event(input, &e);
			} break;
			case SDL_MOUSEBUTTONDOWN:
			case SDL_MOUSEBUTTONUP: {
				SDL_MouseButtonEvent *b = &event.button;
				Event e = {0};
				MouseEvent *m = &e.u.mouse;
				e.kind = event.type == SDL_MOUSEBUTTONDOWN ? EVENT_MOUSE_DOWN : EVENT_MOUSE_UP;
				m->x = b->x;
				m->y = frame.height-1-b->y;
				m->button = MBUTTON_OTHER;
				switch (b->button) {
				case SDL_BUTTON_LEFT: m->button = MBUTTON_LEFT; break;
				case SDL_BUTTON_MIDDLE: m->button = MBUTTON_MIDDLE; break;
				case SDL_BUTTON_RIGHT: m->button = MBUTTON_RIGHT; break;
				}
				add_event(input, &e);
			} break;
			case SDL_MOUSEMOTION: {
				SDL_MouseMotionEvent *motion = &event.motion;
				Event e = {0};
				MouseMoveEvent *m = &e.u.mmove;
				e.kind = EVENT_MOUSE_MOVE;
				m->x = motion->x;
				m->y = frame.height-1-motion->y;
				m->mask = convert_mouse_mask(motion->state);
				add_event(input, &e);
			} break;
			case SDL_TEXTINPUT: {
				Event e = {0};
				TextEvent *t = &e.u.text;
				e.kind = EVENT_TEXT;
				str_cpy(t->text, sizeof t->text, event.text.text);
				add_event(input, &e);
			} break;
			case SDL_MOUSEWHEEL: {
				Event e = {0};
				ScrollEvent *s = &e.u.scroll;
				e.kind = EVENT_SCROLL;
				s->dy = event.wheel.y;
				add_event(input, &e);
			} break;
			}
		}
		{
			Uint8 const *kbd = SDL_GetKeyboardState(NULL);
			u16 i;
			bool *keys_down = input->keys_down;
			for (i = 0; i < NKEYS; ++i) {
				SDL_Scancode scan = key_to_scancode(i);
				keys_down[i] = kbd[scan];
			}
		}
		
		input->focus = focus;
		{
			Uint32 state = SDL_GetMouseState(&input->mouse_x, &input->mouse_y);
			input->mouse_mask = convert_mouse_mask(state);
			input->mouse_y = frame.height-1 - input->mouse_y;
		}
		{
			Uint32 ticks = SDL_GetTicks();
			if (ticks > last_frame_ticks) {
				frame.dt = 0.001 * (ticks - last_frame_ticks);
			} else frame.dt = 0; // the number of ticks wraps around every ~49 days 
			last_frame_ticks = ticks;
		}

		frame.close = input->closed;
		game_frame(&frame);
		if (frame.close) break;
		if (frame.fullscreen != fullscreen) {
			fullscreen = frame.fullscreen;
			SDL_SetWindowFullscreen(window, fullscreen ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);
			if (!fullscreen) {
				SDL_SetWindowSize(window, dfl_windowed_width, dfl_windowed_height);
				SDL_SetWindowPosition(window, 20, 20);
			}
		}
		SDL_GL_SwapWindow(window);
	}

	SDL_DestroyWindow(window);

	return 0;
}
