#ifdef DEBUG
#include "game.h"
#else
#include "game.c"
#endif

#include <X11/Xlib.h>
#include <X11/extensions/Xrandr.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>
#include <X11/XKBlib.h>
#include <GL/glx.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <libgen.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/random.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>

#include "sockets-posix.c"
#include "filesystem-posix.c"
#include "procs.c"

static void die(char const *fmt, ...) {
	va_list args;
	fprintf(stderr, "Error: ");
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(-1);
}

static Bool event_predicate(Display *dpy, XEvent *event, XPointer arg) {
	(void)dpy; (void)event; (void)arg;
	return True;
}

static bool is_key_down(u8 *keys, KeyCode kc) {
	int entry = kc >> 3;
	int bit   = kc & 7;
	return (keys[entry] & (1<<bit)) != 0;
}

static u16 keysym_to_key(KeySym sym) {
	switch (sym) {
	case XK_Left: return KEY_LEFT;
	case XK_Right: return KEY_RIGHT;
	case XK_Down: return KEY_DOWN;
	case XK_Up: return KEY_UP;
	case XK_space: return KEY_SPACE;
	case XK_Control_L: return KEY_LCTRL;
	case XK_Control_R: return KEY_RCTRL;
	case XK_Shift_L: return KEY_LSHIFT;
	case XK_Shift_R: return KEY_RSHIFT;
	case XK_Return: return KEY_ENTER;
	case XK_BackSpace: return KEY_BACKSPACE;
	case XK_Escape: return KEY_ESCAPE;
	case XK_Page_Down: return KEY_PAGEDOWN;
	case XK_Page_Up: return KEY_PAGEUP;
	default:
		if (sym >= XK_a && sym <= XK_z) {
			return (u16)(sym - XK_a + KEY_A);
		} else if (sym >= XK_A && sym <= XK_Z) {
			return (u16)(sym - XK_A + KEY_A);
		} else if (sym >= XK_0 && sym <= XK_9) {
			return (u16)(sym - XK_0 + KEY_0);
		} else if (sym >= XK_F1 && sym <= XK_F12) {
			return (u16)(sym - XK_F1 + KEY_F1);
		}
		break;
	}
	return KEY_UNKNOWN;
}
static KeySym key_to_keysym(u16 key) {
	switch (key) {
	case KEY_LEFT: return XK_Left;
	case KEY_RIGHT: return XK_Right;
	case KEY_DOWN: return XK_Down;
	case KEY_UP: return XK_Up;
	case KEY_SPACE: return XK_space;
	case KEY_LCTRL: return XK_Control_L;
	case KEY_RCTRL: return XK_Control_R;
	case KEY_LSHIFT: return XK_Shift_L;
	case KEY_RSHIFT: return XK_Shift_R;
	case KEY_ENTER: return XK_Return;
	case KEY_BACKSPACE: return XK_BackSpace;
	case KEY_ESCAPE: return XK_Escape;
	case KEY_PAGEDOWN: return XK_Page_Down;
	case KEY_PAGEUP: return XK_Page_Up;
	default:
		if (key >= KEY_A && key <= KEY_Z) {
			return (KeySym)(key - KEY_A + XK_a);
		} else if (key >= KEY_0 && key <= KEY_9) {
			return (KeySym)(key - KEY_0 + XK_0);
		} else if (key >= KEY_F1 && key <= KEY_F12) {
			return (KeySym)(key - KEY_F1 + XK_F1);
		}
		break;
	}
	return 0;
}
#define DFL_WINDOWED_WIDTH 1280
#define DFL_WINDOWED_HEIGHT 720

// semi-borrowed from SDL2 
static void set_window_fullscreen(Display *dpy, Window win, bool fullscreen) {
	Atom wm_state = XInternAtom(dpy, "_NET_WM_STATE", False);
	Atom wm_state_fullscreen = XInternAtom(dpy, "_NET_WM_STATE_FULLSCREEN", False);
	XEvent e = {0};
	XWindowAttributes attribs = {0};
	XGetWindowAttributes(dpy, win, &attribs);
#if 0
	XGetWMNormalHints(dpy, win, &hints, &flags);
	if (fullscreen) {
		// make window resizable 
		hints.flags &= ~(PMinSize | PMaxSize);
	} else {
		// make window non-resizable 
		hints.flags |= PMinSize | PMaxSize;
		hints.min_width  = hints.max_width  = DFL_WINDOWED_WIDTH;
		hints.min_height = hints.max_height = DFL_WINDOWED_HEIGHT;
	}
	XSetWMNormalHints(dpy, win, &hints);
#endif

	e.xany.type = ClientMessage;
	e.xclient.message_type = wm_state;
	e.xclient.format = 32;
	e.xclient.window = win;
	e.xclient.data.l[0] = fullscreen;
	e.xclient.data.l[1] = (long)wm_state_fullscreen;
	e.xclient.data.l[3] = 0;

	XSendEvent(dpy, RootWindowOfScreen(attribs.screen), 0,
		SubstructureNotifyMask | SubstructureRedirectMask, &e);
}

static void add_event(Input *input, Event *e) {
	if (input->nevents < MAX_EVENTS_PER_FRAME) {
		input->events[input->nevents++] = *e;
	}
}

static u8 convert_x_mouse_mask(uint mask) {
	u8 mouse_mask = 0;
	if (mask & Button1Mask) mouse_mask |= MBUTTON_MASK_LEFT;
	if (mask & Button2Mask) mouse_mask |= MBUTTON_MASK_MIDDLE;
	if (mask & Button3Mask) mouse_mask |= MBUTTON_MASK_RIGHT;
	return mouse_mask;
}

int main(int argc, char **argv) {
	Display *dpy = XOpenDisplay(NULL);
	GLint glx_attrs[] = {GLX_RGBA, GLX_DEPTH_SIZE, 24, GLX_DOUBLEBUFFER, None};
	XVisualInfo *visual_info = NULL;
	Window win = 0, root = 0;
	Colormap cmap = 0;
	XSetWindowAttributes win_attrs = {0};
	Atom wm_delete_window = 0;
	GLXContext glx_ctx = NULL;
	int width = 0, height = 0;
	bool fullscreen = true;
	XIC xic = 0;
	XIM xim = 0;

	if (argc) { // make sure we're in the program's directory 
		char *program = argv[0];
		char *dir = dirname(strdup(program));
		if (strcmp(dir, ".") != 0) {
			chdir(dir);
		}
	}

	{
		uint seed;
	#ifdef __linux__ // getrandom isn't in POSIX
		if (getrandom(&seed, sizeof seed, 0) == -1) {
	#endif
			struct timespec t;
			clock_gettime(CLOCK_MONOTONIC, &t);
			seed = (uint)(t.tv_sec + t.tv_nsec);
	#ifdef __linux__
		}
	#endif
		srand(seed);
	}

	if (!dpy)
		die("Couldn't open display.");

#ifdef DEBUG
	width  = DFL_WINDOWED_WIDTH;
	height = DFL_WINDOWED_HEIGHT;
#else
	{
		int nsizes = 0;
		XRRScreenSize *sizes = XRRSizes(dpy, DefaultScreen(dpy), &nsizes);
		width = sizes[0].width;
		height = sizes[0].height;
	}
#endif
	
	root = DefaultRootWindow(dpy);
	visual_info = glXChooseVisual(dpy, 0, glx_attrs);
	if (!visual_info)
		die("Couldn't choose GLX visual.");
	
	cmap = XCreateColormap(dpy, root, visual_info->visual, AllocNone);
	win_attrs.colormap = cmap;
	win_attrs.event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask 
		| FocusChangeMask | PointerMotionMask | StructureNotifyMask;
	win = XCreateWindow(dpy, root, 0, 0, (uint)width, (uint)height, 0, visual_info->depth, InputOutput, visual_info->visual,
		CWColormap | CWEventMask, &win_attrs);
	if (!win)
		die("Couldn't create window.");
	#if 0
	{ // make window non-resizable 
		XSizeHints hints = {0};
		hints.flags = PMinSize | PMaxSize;
		hints.min_width = hints.max_width = width;
		hints.min_height = hints.max_height = height;
		XSetWMNormalHints(dpy, win, &hints);
	}
	#endif
	glx_ctx = glXCreateContext(dpy, visual_info, NULL, GL_TRUE);
	if (!glx_ctx)
		die("Couldn't create GLX context.");
	glXMakeCurrent(dpy, win, glx_ctx);

	wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	if (wm_delete_window)
		XSetWMProtocols(dpy, win, &wm_delete_window, 1);
	XStoreName(dpy, win, "PIONS");

	

	{ // create X input context, which is used to convert key presses to strings 
		xim = XOpenIM(dpy, NULL, NULL, NULL);
		if (!xim)
			die("Couldn't open input method information.");
		xic = XCreateIC(xim, XNClientWindow, win, XNFocusWindow, win,
			XNInputStyle, XIMPreeditNothing | XIMStatusNothing,
			NULL);
		if (!xic)
			die("Couldn't create input context.");
	}

	XMapWindow(dpy, win);
	
	{
		Frame frame = {0};
		Input *input = &frame.input;
		bool focus = true;
		ulong memory_size = frame.memory_size = GAME_MEMORY_REQUIRED;
		void *memory = NULL;
		struct timespec last_frame = {0};
	
	#if DEBUG
		void *so_handle = NULL;
		char so_filename[32] = {0};
		void (*game_frame)(Frame *) = NULL;
		struct timespec so_last_mtime = {0};
		void *requested_address = (void *)(1UL << 30);
		memory = mmap(requested_address, memory_size, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
		if (memory != requested_address)
			die("Couldn't get requested address.");
	#else
		memory = calloc(1, memory_size);
		if (!memory)
			die("Couldn't get enough memory to run.");
	#endif
		
		frame.fullscreen = fullscreen;
		frame.memory = memory;

		fill_out_procs(&frame.procs, (void (*(*)(char const *))(void))glXGetProcAddress);

		clock_gettime(CLOCK_MONOTONIC, &last_frame);

		while (1) {
			bool window_mapped_this_frame = false;
		#if DEBUG
			// check for so changes 
			{
				struct stat statbuf = {0};
				struct timespec mtime;
				stat("./obj/game.so_changed", &statbuf);
				mtime = statbuf.st_mtim;
				if (mtime.tv_sec != so_last_mtime.tv_sec || mtime.tv_nsec != so_last_mtime.tv_nsec) {
					char filename[32], command[64];
					
					so_last_mtime = mtime;

					sprintf(filename, "./obj/game%x%x.so", rand(), rand());
					sprintf(command, "cp ./obj/game.so %s", filename);
					if (system(command) == 0) {
						void *new_handle = dlopen(filename, RTLD_NOW);
						if (new_handle) {
						no_warn_start
							void (*sym)(Frame *) = (void (*)(Frame *))dlsym(new_handle, "game_frame");
						no_warn_end
							if (sym) {
								if (so_handle) dlclose(so_handle);
								so_handle = new_handle;
								game_frame = sym;
								strcpy(so_filename, filename);
								printf("Successfully loaded dynamic library %s.\n", filename);
								{ // delete old .so files.
									DIR *dir = opendir("obj");
									if (dir) {
										struct dirent *ent = NULL;
										while ((ent = readdir(dir))) {
											if (ent->d_type == DT_REG) {
												char *name = ent->d_name;
												// check if it's game<something>.so
												if (strlen(name) > 7 &&
													strncmp(name, "game", 4) == 0 && 
													strcmp(&name[strlen(name)-3], ".so") == 0) {
													// make sure it's not the .so file we're using right now
													if (strcmp(name, so_filename + strlen("./obj/")) != 0) {
														char full_name[256] = {0};
														snprintf(full_name, sizeof full_name, "obj/%s", name);
														printf("Deleting old .so file %s.\n", full_name);
														unlink(full_name); // delete it
													}
												}
											}
										}
										closedir(dir);
									}
								}
							} else {
								dlclose(new_handle);
								unlink(filename);
								fprintf(stderr, "!!! Couldn't get game_frame: %s\n", dlerror());
							}
						} else {
							unlink(filename);
							fprintf(stderr, "!!! Failed to open dynamic library: %s\n", dlerror());
						}
					} else {
						fprintf(stderr, "!!! Failed to copy dynamic library.\n");
					}
				}
			}
			if (!game_frame)
				die("Couldn't get game_frame.");
		#endif
			
			memset(input, 0, sizeof *input);
			{
				XEvent event = {0};
				while (XCheckIfEvent(dpy, &event, event_predicate, NULL)) {
					switch (event.type) {
					case MapNotify:
						window_mapped_this_frame = true;
						break;
					case ClientMessage:
						if ((Atom)event.xclient.data.l[0] == wm_delete_window)
							input->closed = true;
						break;
					case KeyRelease: {
						KeySym sym = XkbKeycodeToKeysym(dpy, (KeyCode)event.xkey.keycode, 0, (event.xkey.state & ShiftMask) != 0);
						Event e = {0};
						KeyEvent *k = &e.u.key;
						e.kind = EVENT_KEY_UP;
						k->key = keysym_to_key(sym);
					} break;
					case KeyPress: {
						Status status = 0;
						KeySym sym = 0;
						Event ekey = {0}; // key event 
						Event etext = {0}; // text event 
						bool has_key = false, has_text = false;
						KeyEvent *k = &ekey.u.key;
						TextEvent *t = &etext.u.text;
						ekey.kind = EVENT_KEY_DOWN;
						etext.kind = EVENT_TEXT;
						/* 
							convert the key press event to both a KeySym and text.
							it's sizeof k->text - 1 because of the null terminator.
						*/
						Xutf8LookupString(xic, &event.xkey, t->text, sizeof t->text - 1, &sym, &status);
						if (status == XLookupChars || status == XLookupBoth)
							has_text = true;
						if (status == XLookupKeySym || status == XLookupBoth)
							has_key = true;
						if (has_text)
							add_event(input, &etext);
						if (has_key) {
							k->key = keysym_to_key(sym);
							add_event(input, &ekey);
						}
					} break;
					case ButtonPress:
					case ButtonRelease: {
						XButtonEvent *button = &event.xbutton;
						Event e = {0};
						u8 mbutton = MBUTTON_OTHER;
						int scroll_dy = 0;
						switch (button->button) {
						case Button1:
							mbutton = MBUTTON_LEFT;
							break;
						case Button2:
							mbutton = MBUTTON_MIDDLE;
							break;
						case Button3:
							mbutton = MBUTTON_RIGHT;
							break;
						case Button4: // Button4 is scroll up
							scroll_dy = +1;
							break;
						case Button5: // Button5 is scroll down
							scroll_dy = -1;
							break;
						}
						if (mbutton != MBUTTON_OTHER) {
							MouseEvent *m = &e.u.mouse;
							e.kind = event.type == ButtonPress ? EVENT_MOUSE_DOWN : EVENT_MOUSE_UP;
							m->button = mbutton;
							m->x = button->x;
							m->y = frame.height-1-button->y;
							add_event(input, &e);
						}
						if (scroll_dy) {
							ScrollEvent *s = &e.u.scroll;
							e.kind = EVENT_SCROLL;
							s->dy = scroll_dy;
							add_event(input, &e);
						}
					} break;
					case MotionNotify: {
						XMotionEvent *motion = &event.xmotion;
						Event e = {0};
						MouseMoveEvent *m = &e.u.mmove;
						e.kind = EVENT_MOUSE_MOVE;
						m->x = motion->x;
						m->y = frame.height-1-motion->y;
						m->mask = convert_x_mouse_mask(motion->state);
						add_event(input, &e);
					} break;
					case FocusIn:
						focus = true;
						break;
					case FocusOut:
						focus = false;
						break;
					}
				}
			}
			input->focus = focus;
			if (focus) { // check which keys are down 
				u8 keys[32] = {0};
				bool *keys_down = input->keys_down;
				u16 i;
				XQueryKeymap(dpy, (char *)keys);
				for (i = 0; i < NKEYS; ++i) {
					KeyCode code = XKeysymToKeycode(dpy, key_to_keysym(i));
					keys_down[i] = is_key_down(keys, code);
				}

			}
			{
				XWindowAttributes attrs = {0};
				int mouse_x = 0, mouse_y = 0;
				int int_unused;
				Window win_unused;
				uint mask = 0;
				XGetWindowAttributes(dpy, win, &attrs);
				frame.width = attrs.width, frame.height = attrs.height;
				XQueryPointer(dpy, win, &win_unused, &win_unused, &int_unused, &int_unused, &mouse_x, &mouse_y, &mask);
				mouse_y = frame.height-1 - mouse_y;

				input->mouse_x = mouse_x;
				input->mouse_y = mouse_y;
				input->mouse_mask = convert_x_mouse_mask(mask);
			}
			
			{
				struct timespec this_frame = {0};
				double dt = 0;
				clock_gettime(CLOCK_MONOTONIC, &this_frame);
				dt = (double)(this_frame.tv_sec - last_frame.tv_sec)
					+ (double)(this_frame.tv_nsec - last_frame.tv_nsec) / 1000000000.0;
				last_frame = this_frame;
				frame.dt = dt;
				frame.close = input->closed;
				game_frame(&frame);
				if (frame.close) break;
				if (frame.fullscreen != fullscreen) {
					fullscreen = frame.fullscreen;
				#if DEBUG
					printf("%s fullscreen.\n", fullscreen ? "Enabling" : "Disabling");
				#endif
					set_window_fullscreen(dpy, win, fullscreen);
				} else if (window_mapped_this_frame) {
					// set window fullscreen only works when the window is mapped 
					set_window_fullscreen(dpy, win, fullscreen);
				}
				#if 0
				if (frame.resize) {
					XSizeHints hints = {0};
				#if DEBUG
					printf("Resize to %d,%d\n",frame.width,frame.height);
				#endif
					hints.flags = PMinSize | PMaxSize;
					hints.min_width = hints.max_width = width = frame.width;
					hints.min_height = hints.max_height = height = frame.height;
					XSetWMNormalHints(dpy, win, &hints);
					XResizeWindow(dpy, win, (uint)width, (uint)height);
				}
			#endif
			}

			glXSwapBuffers(dpy, win);
		}

	#if DEBUG
		if (*so_filename) unlink(so_filename);
	#endif
	}

	return 0;
}
