/*
get map names which match the search term `matching`. the number of maps is returned in nmaps, if non-NULL.
*/
static char **get_maps(State *state, char const *matching, int *nmaps) {
	PlatformProcs *procs = &state->procs;
	char **files = procs->list_files("maps");
	if (files) {
		int in, out = 0, nfiles = 0;
		for (in = 0; files[in]; ++in) {
			if (stristr(files[in], matching)) {
				files[out++] = files[in];
			} else {
				free(files[in]);
				files[in] = NULL;
			}
		}
		files[out] = NULL; // make sure resulting array is NULL-terminated 
		nfiles = out;
		if (nfiles > 0)
			qsort(files, (size_t)nfiles, sizeof *files, qsort_stricmp); // sort maps alphabetically 
		if (nmaps) *nmaps = nfiles;
	} else {
		if (nmaps) *nmaps = 0;
	}
	return files;
}

static void free_maps(char **maps) {
	int i;
	if (maps) {
		for (i = 0; maps[i]; ++i)
			free((void *)maps[i]);
		free((void *)maps);
	}
}

/*
returns:
1 - successful load
0 - map doesn't exist
-1 - some other error
*/
static int map_load(State *state, Game *game, char const *name, bool display_errors, GameScreen return_to_screen) {
	u32 mark = tmp_push(state);
	int ret = -1;

	if (*name) {
		char filename[512] = "maps/";
		FILE *fp = NULL;
		size_t file_size = 0;
		u8 *data = NULL;

		strncat(filename, name, 500);
		fp = fopen(filename, "rb");
		if (fp) {
			fseek(fp, 0, SEEK_END);
			file_size = (size_t)ftell(fp);
			fseek(fp, 0, SEEK_SET);
			if (file_size < MAP_MAX_FILE_SIZE && (data = tmp_alloc(state, file_size))) {
				if (fread(data, 1, file_size, fp) == file_size) {
					ReadBuffer b = mkrbuffer(data, file_size);
					GameKind game_kind = game->kind;
					buf_read_game_setup(&b, state, game);
					if (!b.error) {
						game->kind = game_kind;
						ret = 1;
					} else {
						if (display_errors)
							display_error(state, return_to_screen, false, "Error loading map: %s", b.error_message);
					}
				} else {
					if (display_errors)
						display_error(state, return_to_screen, false, "Error reading map file.");
				}
			} else {
				if (display_errors)
					display_error(state, return_to_screen, false, "Map too %s (size: %lu).",
						file_size > 0 ? "large" : "small", (ulong)file_size);
			}
			fclose(fp);
		} else {
			// map doesn't exist 
			ret = 0;
		}
	}
	tmp_pop(state, mark);
	return ret;
}

static void map_selector_scrollbar_handle_mouse(State *state, int nmaps, v2 mpos, Rect scrollbar, Rect scroller) {
	MapSelector *mapsel = &state->mapsel;
	float p;
	mpos.y += scroller.size.y * 0.5f;
	p = 1-normf(mpos.y, rect_y1(scrollbar), rect_y2(scrollbar)); // 0 = top, 1 = bottom
	mapsel->scroll = (int)((float)nmaps * p);
}

/*
	Display a map selection menu. Arguments:
	search - if non-NULL, text input used to narrow down results.
	bottom_y - y position of the bottom of the selector
	top_y - starting (top) y position of the menu
	menu_width - the full width of the menu
	selected_map - if a map was selected, it's returned here. otherwise it'll be an empty string.

	returns
	-1 if no map was selected
	 0 if a pre-existing map was selected
	 1 if a new map was selected (by pressing enter while focus is on the search box)
*/
static int map_selector(Frame *frame, TextInput *search, float bottom_y, float top_y, float menu_width, char selected_map[MAP_NAME_BUFSZ]) {
	State *state = frame_state(frame);
	Input *input = &frame->input;
	Event *events = input->events, *events_end = events + input->nevents;
	float width_gl = state->width_gl;
	float center_x = width_gl * 0.5f;
	float x1 = center_x - 0.5f * menu_width, x2 = center_x + 0.5f * menu_width; // left and right sides of the menu
	float end_y; // the menu might end before bottom_y, because we always show an integer number of options.
	v2 mouse_pos = state->mouse_pos;
	MapSelector *mapsel = &state->mapsel;
	int maps_shown = 0;
	int ret = -1;
	int scroll = (int)mapsel->scroll;
	TextFont *font = &state->main_font;
	float scrollbar_width = 0.02f;
	bool show_scrollbar = false;
	Rect scrollbar; // rectangle taken up by scrollbar
	Rect scroller; // rectangle for the thing on the scroll bar
	float spacing_y = 1.5f * font->char_height_gl; // vertical spacing between items in the menu.

	maps_shown = ((int)((top_y - bottom_y) / spacing_y));
 	end_y = top_y - spacing_y * (float)maps_shown;

	memset(selected_map, 0, MAP_NAME_BUFSZ);

	text_input_render(search, state);
	{
		int i, map_idx, nmaps = 0;
		int hovering_over = -1; // which map are we hovering over?
		
		char **maps = get_maps(state, search ? search->text : "", &nmaps);

		if (scroll > nmaps - maps_shown) {
			scroll = mapsel->scroll = nmaps - maps_shown;
		}
		if (scroll < 0) scroll = mapsel->scroll = 0;

		if (nmaps == 0) {
			gl_color1f(0.5f);
			font_begin(font);
			text_render_centered(font, "No results.", V2(center_x, top_y));
			font_end();
		} else {
			if (nmaps <= maps_shown) { 
				// no scrollbar needed; all results will appear
				maps_shown = nmaps;
			} else {
				// many results; show scrollbar
				show_scrollbar = true;
				x2 -= scrollbar_width;
				scrollbar = rect4(x2, end_y, x2+scrollbar_width, top_y + gl_pixel(state));
			}

			if (show_scrollbar) { 
				// calculate position and height of scroller
				float ptop = (float)scroll/(float)nmaps,
					pbottom = (float)(scroll+maps_shown)/(float)nmaps;
				float scroller_top    = lerpf(ptop,    top_y, end_y);
				float scroller_bottom = lerpf(pbottom, top_y, end_y);
				scroller = rect4(x2, scroller_bottom, x2+scrollbar_width, scroller_top + gl_pixel(state));
				
				glBegin(GL_QUADS);
				// render scrollbar
				gl_color1f((mouse_in_rect(state, scrollbar) || mapsel->scrolling) ? 0.22f : 0.2f);
				rect_render(scrollbar);
				// render scroller
				{
					float scroller_color = 0.6f;
					if (mouse_in_rect(state, scroller) || mapsel->scrolling) {
						if ((state->mouse_mask & MBUTTON_MASK_LEFT) || mapsel->scrolling)
							scroller_color = 0.5f;
						else
							scroller_color = 0.75f;
					}
					gl_color1f(scroller_color);
					rect_render(scroller);
				}
				glEnd();
			}

			glBegin(GL_LINES);
			for (i = 0; i <= maps_shown; ++i) {
				float xm = (x1 + x2) * 0.5f;
				float y = top_y - spacing_y*(float)i;
				float y1 = y - spacing_y + gl_pixel(state), y2 = y;
				Rect bbox = rect4(x1, y1, x2, y2);

				map_idx = i + scroll;
				if (map_idx > nmaps) break;

				assert(y >= bottom_y);

				// lines separating menu items 
				gl_color1f(0.2f);
				glVertex2f(x1, y);
				gl_color1f(0.8f);
				glVertex2f(xm, y);
				glVertex2f(xm, y);
				gl_color1f(0.2f);
				glVertex2f(x2, y);

				if (i < maps_shown) {
					Event *e;
					if (rect_contains_point(bbox, mouse_pos)) {
						v3 color = V3(0,0,1);
						hovering_over = map_idx;
						// highlight the item we're hovering over 
						glEnd();
						glBegin(GL_QUADS);
						v3_gl_color_alpha(color, 0);
						glVertex2f(x1, y1);
						glVertex2f(x1, y2);
						v3_gl_color_alpha(color, 0.4f);
						glVertex2f(xm, y2);
						glVertex2f(xm, y1);
						glVertex2f(xm, y1);
						glVertex2f(xm, y2);
						v3_gl_color_alpha(color, 0);
						glVertex2f(x2, y2);
						glVertex2f(x2, y1);
						glEnd();
						glBegin(GL_LINES);
					}
					for (e = events; e != events_end; ++e) {
						if (e->kind == EVENT_MOUSE_DOWN) {
							MouseEvent *m = &e->u.mouse;
							v2 mpos = pixels_to_gl_coords(state, m->x, m->y);
							if (m->button == MBUTTON_LEFT && rect_contains_point(bbox, mpos)) {
								str_cpy(selected_map, MAP_NAME_BUFSZ, maps[map_idx]);
								ret = 0;
							}
						}
					}
				}
			}
			glEnd();

			font_begin(font);
			if (nmaps > 0) {
				// render map names 
				for (i = 0; i < maps_shown; ++i) {
					char str[MAP_NAME_BUFSZ] = {0};
					float y = top_y - spacing_y*((float)i + 0.5f);

					map_idx = i + scroll;
					
					if (map_idx >= 0 && map_idx < nmaps) {
						str_cpy(str, sizeof str, maps[map_idx]);
						gl_color_text();
						text_render_centered(font, str, V2(center_x, y));
					}
				}
			}
			font_end();

			if (hovering_over != -1) {
				Game *game_preview = &state->game_preview;
				if (map_load(state, game_preview, maps[hovering_over], false, SCR_MAIN_MENU) == 1) {
					Settings *settings = &state->settings;
					if (settings->show_map_previews) {
						// render semi-transparent preview of the map the user is hovering over
						hex_setup_game_grid(game_preview, V2(0,0), V2(width_gl, 1));
						render_game_board(state, game_preview, 0.3f, false);
					}
				}
			}
		}

		if (!(state->mouse_mask & MBUTTON_MASK_LEFT)) {
			// this is to handle the case where the user clicks on the scrollbar,
			// and presses escape before releasing, then returns to the map selector
			// (we don't want to start out scrolling)
			mapsel->scrolling = false;
		}

		// in event processing, we don't have to worry about mapsel->scroll being in bounds, because it will be
		// corrected next frame (also we don't have to update the scroll variable because it's not used)

		{
			Event *e;
			for (e = events; e != events_end; ++e) {
				if (search && text_input_event(search, state, e)) {
					str_cpy(selected_map, MAP_NAME_BUFSZ, search->text);
					ret = 1; // new map 
					for (i = 0; i < nmaps; ++i) {
						if (strcmp(maps[i], selected_map) == 0)
							ret = 0; // pre-existing map 
					}
				}
				switch (e->kind) {
				case EVENT_KEY_DOWN:
					if (show_scrollbar) {
						// use keys to scroll
						switch (e->u.key.key) {
						case KEY_PAGEDOWN:
							mapsel->scroll += maps_shown;
							break;
						case KEY_PAGEUP:
							mapsel->scroll -= maps_shown;
							break;
						case KEY_UP:
							--mapsel->scroll;
							break;
						case KEY_DOWN:
							++mapsel->scroll;
							break;
						}
					}
					break;
				case EVENT_SCROLL:
					if (show_scrollbar)
						mapsel->scroll -= clampi(e->u.scroll.dy, -1, +1);
					break;
				case EVENT_MOUSE_MOVE: {
					MouseMoveEvent *m = &e->u.mmove;
					v2 pos = pixels_to_gl_coords(state, m->x, m->y);
					if (show_scrollbar) {
						// drag mouse to scroll
						if (state->mouse_mask & MBUTTON_MASK_LEFT) {
							if (mapsel->scrolling) {
								map_selector_scrollbar_handle_mouse(state, nmaps, pos, scrollbar, scroller);
							}
						}
					}
				} break;
				case EVENT_MOUSE_DOWN: {
					MouseEvent *m = &e->u.mouse;
					v2 pos = pixels_to_gl_coords(state, m->x, m->y);
					if (show_scrollbar) {
						if (m->button == MBUTTON_LEFT) {
							if (rect_contains_point(scrollbar, pos)) {
								map_selector_scrollbar_handle_mouse(state, nmaps, pos, scrollbar, scroller);
								mapsel->scrolling = true;
							}
						}
					}
				} break;
				case EVENT_MOUSE_UP: {
					MouseEvent *m = &e->u.mouse;
					if (m->button == MBUTTON_LEFT) {
						mapsel->scrolling = false;
					}
				} break;
				}
			}
		}

		free_maps(maps);
	}

	return ret;
}

// if the maps directory does not exist, create the maps directory and copy if  
static void copy_starting_maps_if_needed(State *state) {
	PlatformProcs *procs = &state->procs;
	char **starting_map_files = NULL;
	int i, nfiles = 0;

	if (procs->mkdir("maps") != 1) {
		// either the directory already exists, or we couldn't create it 
		return;
	}
	starting_map_files = procs->list_files("starting_maps");
	if (starting_map_files) {
		for (i = 0; starting_map_files[i]; ++i);
		nfiles = i;
	}

	for (i = 0; i < nfiles; ++i) {
		char const *map_name = starting_map_files[i];
		char in_name[64] = "starting_maps/", out_name[64] = "maps/";
		str_cat(in_name,  sizeof in_name,  map_name);
		str_cat(out_name, sizeof out_name, map_name);
		debug_log("Copying map %s => %s\n", in_name, out_name);
		copy_file(in_name, out_name); // if this fails there's not much we can do 
	}

	for (i = 0; i < nfiles; ++i)
		free(starting_map_files[i]);

}
