static void sprite_create(Texture *tex, Sprite *sprite, int x, int y, int w, int h) {
	float x_scale = 1.0f / (float)tex->width,
		y_scale = 1.0f / (float)tex->height;
	sprite->x = (float)x * x_scale;
	sprite->w = (float)w * x_scale;
	sprite->y = (float)y * y_scale;
	sprite->h = (float)h * y_scale;
	sprite->pix_w = (u16)w;
	sprite->pix_h = (u16)h;
}

static void sprite_render(State *state, SpriteID sprite_id, v2 pos, v2 size) {
	Sprite *sprite = state->sprites + sprite_id;
	float x1 = pos.x, y1 = pos.y;
	float x2 = x1+size.x, y2 = y1+size.y;

	float s1 = sprite->x, t1 = sprite->y;
	float s2 = s1+sprite->w, t2 = t1+sprite->h;
	glTexCoord2f(s1, t1); glVertex2f(x1, y1);
	glTexCoord2f(s1, t2); glVertex2f(x1, y2);
	glTexCoord2f(s2, t2); glVertex2f(x2, y2);
	glTexCoord2f(s2, t1); glVertex2f(x2, y1);
}

// draw sprite rotated by a multiple of 90 degrees (left is considered the default direction) 
static void sprite_render_direction(State *state, SpriteID sprite_id, Direction dir, v2 pos, v2 size) {
	Sprite *sprite = state->sprites + sprite_id;
	float x1 = pos.x, y1 = pos.y;
	float x2 = x1+size.x, y2 = y1+size.y;

	float s1 = sprite->x, t1 = sprite->y;
	float s2 = s1+sprite->w, t2 = t1+sprite->h;

	v2 points[4];
	points[0] = V2(x1,y1);
	points[1] = V2(x1,y2);
	points[2] = V2(x2,y2);
	points[3] = V2(x2,y1);
	
	glTexCoord2f(s1, t1); v2_gl_vertex(points[dir]);
	glTexCoord2f(s1, t2); v2_gl_vertex(points[(dir+1)%4]);
	glTexCoord2f(s2, t2); v2_gl_vertex(points[(dir+2)%4]);
	glTexCoord2f(s2, t1); v2_gl_vertex(points[(dir+3)%4]);
}

static void sprite_render_centered(State *state, SpriteID sprite_id, v2 center, v2 size) {
	sprite_render(state, sprite_id, v2_sub(center, v2_scale(size, 0.5f)), size);
}

static void sprite_render_direction_centered(State *state, SpriteID sprite_id, Direction dir, v2 center, v2 size) {
	sprite_render_direction(state, sprite_id, dir, v2_sub(center, v2_scale(size, 0.5f)), size);
}

static void load_bitmap(State *state, char const *filename, Texture *texture) {
	GLProcs *gl = &state->gl;
	FILE *file = fopen(filename, "rb");
	memset(texture, 0, sizeof *texture);
#if DEBUG
	strcpy(texture->filename, filename);
#endif
	if (file) {
		int bm1 = getc(file);
		int bm2 = getc(file);
		#if DEBUG
		struct stat statbuf;
		if (stat(filename, &statbuf) == 0) {
			// NOTE: we're not using st_mtim here because it's not supported on windows 
			texture->last_mtime = statbuf.st_mtime;
		}
		#endif

		if (bm1 != 'B' || bm2 != 'M') {
			printf("!!! %s is not a BMP file.\n", filename);
		} else {
			u32 data_offset, width, height, words[7] = {0};
			u16 bits_per_pixel;
			// read some 32-bit words 
			fread(&words, sizeof *words, arr_count(words), file);
			// words[0] is the size of the file, words[1] is just reserved stuff 
			data_offset = words[2];
			// words[3] is the header size 
			width = words[4]; height = words[5];
			texture->width = (int)width;
			texture->height = (int)height;
			bits_per_pixel = (u16)(words[6] >> 16);
			if (bits_per_pixel != 32) {
				printf("!!! Expected 32 bits per pixel in bitmap %s, but got %u.\n", filename, bits_per_pixel);
			} else {
				u32 *bitmap = state->img_mem;
				if (fseek(file, (long)data_offset, SEEK_SET) != 0) {
					printf("!!! Failed to seek to image data in %s.\n", filename);
				} else {
					u32 area = width * height;
					if (fread(bitmap, sizeof *bitmap, area, file) != area) {
						printf("!!! Failed to read image in %s.\n", filename);
					} else {
						{ // create opengl texture 
							GLuint tex;
							bool mipmaps = gl->GenerateMipmap != NULL;
							glGenTextures(1, &tex);
							texture->tex = tex;
							glBindTexture(GL_TEXTURE_2D, tex);
							glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, (GLsizei)width, (GLsizei)height, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, bitmap);
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
							glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
							if (mipmaps) gl->GenerateMipmap(GL_TEXTURE_2D);
							glBindTexture(GL_TEXTURE_2D, 0);
						#if DEBUG
							printf("Loaded bitmap %s.\n", filename);
						#endif
						}
					}
				}
			}
		}
		fclose(file);
	} else {
		printf("!!! Image %s missing.\n", filename);
	}
}

#if DEBUG
static void bitmap_check_for_changes(State *state, Texture *t) {
	struct stat statbuf = {0};
	if (stat(t->filename, &statbuf) == 0) {
		time_t mtime = statbuf.st_mtime, last = t->last_mtime;
		if (mtime != last) {
			Texture prev = *t;
			t->tex = 0;
			load_bitmap(state, prev.filename, t);
			if (t->tex) {
				// load successful; delete old texture 
				glDeleteTextures(1, &prev.tex);
			} else {
				// load unsuccessful; revert to old texture 
				*t = prev;
			}
		}
	} else {
		printf("!!! File missing: %s\n", t->filename);
	}
}
#endif

static void gl_bind_texture(Texture *tex) {
	glBindTexture(GL_TEXTURE_2D, tex->tex);
}

static GLuint compile_shader(State *state, char const *code, GLenum shader_type) {
	GLProcs *gl = &state->gl;
	GLuint shader = gl->CreateShader(shader_type);
	GLint status = 0;

	gl->ShaderSource(shader, 1, &code, NULL);
	gl->CompileShader(shader);
	gl->GetShaderiv(shader, GL_COMPILE_STATUS, &status);
	if (status == GL_FALSE) {
		char log[256] = {0};
		gl->GetShaderInfoLog(shader, sizeof log - 1, NULL, log);
		printf("Error compiling shader: %s.", log);
		return 0;
	}
	return shader;
}

static GLuint compile_shader_from_file(State *state, char const *filename, GLenum shader_type) {
	FILE *fp = fopen(filename, "rb");
	static char contents[16384];
	size_t len;
	if (!fp) {
		printf("Couldn't open shader file: %s.\n", filename);
		return 0;
	}
	len = fread(contents, 1, sizeof contents, fp);
	if (len == sizeof contents) {
		printf("Shader too long.\n");
		return 0;
	}
	contents[len] = 0;
	return compile_shader(state, contents, shader_type);
}

static GLuint link_program(State *state, GLuint *shaders, size_t nshaders) {
	GLProcs *gl = &state->gl;
	GLuint program = gl->CreateProgram();
	GLint status = 0;
	size_t i;
	for (i = 0; i < nshaders; ++i) {
		if (!shaders[i]) {
			// this shader's compilation already failed 
			gl->DeleteProgram(program);
			return 0;
		}
		gl->AttachShader(program, shaders[i]);
	}
	gl->LinkProgram(program);
	gl->GetProgramiv(program, GL_LINK_STATUS, &status);
	if (status == GL_FALSE) {
		char log[256] = {0};
		gl->GetProgramInfoLog(program, sizeof log - 1, NULL, log);
		printf("Error linking shader: %s.", log);
		gl->DeleteProgram(program);
		return 0;
	}
	for (i = 0; i < nshaders; ++i) {
		gl->DeleteShader(shaders[i]);
		shaders[i] = 0;
	}
	return program;
}

static GLuint compile_and_link_shaders(State *state, char const *vshader_filename, char const *fshader_filename) {
	GLuint shaders[2], program;
	shaders[0] = compile_shader_from_file(state, vshader_filename, GL_VERTEX_SHADER);
	shaders[1] = compile_shader_from_file(state, fshader_filename, GL_FRAGMENT_SHADER);
	program = link_program(state, shaders, 2);
#if DEBUG
	if (program) printf("Successfully linked program (vertex shader = %s, program = %u)\n", vshader_filename, program);
#endif
	return program;
}

static GLuint get_attrib_loc(State *state, GLuint program, char const *attrib) {
	GLProcs *gl = &state->gl;
	GLint loc = gl->GetAttribLocation(program, attrib);
	if (loc == -1) {
		printf("Couldn't find vertex attribute %s.\n", attrib);
		return 0;
	}
	return (GLuint)loc;
}

static GLint get_uniform_loc(State *state, GLuint program, char const *uniform) {

	GLProcs *gl = &state->gl;
	GLint loc = gl->GetUniformLocation(program, uniform);
	if (loc == -1) {
		printf("Couldn't find uniform: %s.\n", uniform);
		return -1;
	}
	return loc;
}


static int load_font_(State *state, TextFont *font, u8 const *data,
	float char_height, int bitmap_width, int bitmap_height) {

	u32 mark = tmp_push(state);
	u8 *bitmap = tmp_alloc(state, (size_t)bitmap_width * (size_t)bitmap_height);
	int ret = 1;

	/*
		The number of bytes in a row of a texture must be a multiple of 4.
		See: https://www.khronos.org/opengl/wiki/Common_Mistakes#Texture_upload_and_pixel_reads
	*/
	assert(bitmap_width % 4 == 0);

	font->res_y = (float)state->win_height;
	font->char_height_gl = char_height;
	font->char_height = char_height *= font->res_y;
	font->tex_width = bitmap_width;
	font->tex_height = bitmap_height;
	font->data = data;

	if (bitmap) {
		int err;
		err = stbtt_BakeFontBitmap(data, 0, char_height, bitmap, bitmap_width, bitmap_height, 32, arr_count(font->char_data),
			font->char_data);
		// printf("%p %d %f %p %d %d %d %lu %p\n",data, 0, char_height, bitmap, bitmap_width, bitmap_height, 32, arr_count(font->char_data), font->char_data); 
		if (err > 0) {
			GLuint texture;
			glGenTextures(1, &texture);
			glBindTexture(GL_TEXTURE_2D, texture);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, bitmap_width, bitmap_height, 0, GL_ALPHA, GL_UNSIGNED_BYTE, bitmap);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			font->texture = texture;
			debug_log("Loaded font (size = %f, res_y = %f), using %d rows of a %dx%d bitmap.\n",
				char_height, font->res_y, err, bitmap_width, bitmap_height);
		} else {
			// bitmap not big enough 
			ret = 0;
		}
	} else {
		printf("!!! Not enough memory for font bitmap.\n");
		ret = -1;
	}
	tmp_pop(state, mark);
	return ret;
}

/* 
data = contents of a ttf file
char_height is in opengl coords (1 = window height)
do NOT free data after loading the font, if you want to use update_font_if_necessary
returns true on success
*/
static bool load_font(State *state, TextFont *font, u8 const *data, float char_height) {
	int bitmap_width = 400, bitmap_height = 400;
	while (1) {
		int r = load_font_(state, font, data, char_height, bitmap_width, bitmap_height);
		if (r == -1) {
			return false;
		} else if (r == 1) {
			break;
		}
		bitmap_width *= 2;
		bitmap_height *= 2;
	}
	return true;
}

static bool reload_font(State *state, TextFont *font, float char_height) {
	if (font->texture) glDeleteTextures(1, &font->texture);
	return load_font(state, font, font->data, char_height);
}

static void update_font_if_necessary(State *state, TextFont *font) {
	int y1 = (int)font->res_y, y2 = state->win_height;
	if (y1 != y2) {
		reload_font(state, font, font->char_height_gl);
	}
}

static void text_render_(TextFont *font, char const *s, float *xp, float *yp, bool render, float clip_left) {
	stbtt_bakedchar *cdata = font->char_data;
	int tex_w = font->tex_width, tex_h = font->tex_height;
	float x = *xp, y = *yp;
	// stb_truetype uses a pixel-based top-is-y=0 coordinate system, so we need to convert back and forth 
	float resolution = font->res_y;
	float inv_resolution = 1 / resolution;
	
	x *= resolution;
	y = 1 - y;
	y *= resolution;

	for (; *s; ++s) {
		char c = *s;
		if (c >= 32 && (size_t)c < 32+arr_count(font->char_data)) {
			stbtt_aligned_quad q = {0};
			stbtt_GetBakedQuad(cdata, tex_w, tex_h, c - 32, &x, &y, &q, 1);
			if (render) {
				q.x0 *= inv_resolution;
				q.y0 *= inv_resolution;
				q.x1 *= inv_resolution;
				q.y1 *= inv_resolution;
				q.y0 = 1 - q.y0;
				q.y1 = 1 - q.y1;
#if 1
				if (q.x1 < clip_left) {
					// entirety of character is to the left of clip_left. 
				} else if (q.x0 < clip_left) {
					// part of character is to the left of clip_left 
					float char_width = q.x1 - q.x0; // full width of character 
					float display_char_width = q.x1 - clip_left; // width of character which is being rendered, after clipping 
					float char_fraction = display_char_width / char_width; // fraction of character rendered 
					float s0 = q.s1 - char_fraction * (q.s1 - q.s0); // calculate where to start from 
					float x0 = clip_left;
					glTexCoord2f(  s0, q.t0); glVertex2f(  x0, q.y0);
					glTexCoord2f(q.s1, q.t0); glVertex2f(q.x1, q.y0);
					glTexCoord2f(q.s1, q.t1); glVertex2f(q.x1, q.y1);
					glTexCoord2f(  s0, q.t1); glVertex2f(  x0, q.y1);
				} else 
#endif
				{
					// no clipping 
					#if 0
					printf("%f %f %f %f\n",q.s0*tex_w,q.t0*tex_h,q.s1*tex_w,q.t1*tex_h);
					#endif
					glTexCoord2f(q.s0, q.t0); glVertex2f(q.x0, q.y0);
					glTexCoord2f(q.s1, q.t0); glVertex2f(q.x1, q.y0);
					glTexCoord2f(q.s1, q.t1); glVertex2f(q.x1, q.y1);
					glTexCoord2f(q.s0, q.t1); glVertex2f(q.x0, q.y1);
				}
			}
		}
	}
	*xp =     x * inv_resolution;
	*yp = 1 - y * inv_resolution;
}

static void text_render2f(TextFont *font, char const *s, float x, float y) {
	text_render_(font, s, &x, &y, true, 0);
}
static void text_render(TextFont *font, char const *s, v2 pos) {
	text_render_(font, s, &pos.x, &pos.y, true, 0);
}

static float font_char_height(TextFont *font) {
	return font->char_height_gl * 0.666667f;
}

static v2 get_text_size(TextFont *font, char const *s) {
	float x = 0, y = 0;
	text_render_(font, s, &x, &y, false, 0);
	return V2(x, y + font_char_height(font));
}

static void text_render_centered(TextFont *font, char const *s, v2 center) {
	text_render(font, s, v2_sub(center, v2_scale(get_text_size(font, s), 0.5f)));
}

static void free_texture(State *state, Texture *texture) {
	GLProcs *gl = &state->gl;
	gl->DeleteTextures(1, &texture->tex);
}
static void free_font(State *state, TextFont *font) {
	GLProcs *gl = &state->gl;
	gl->DeleteTextures(1, &font->texture);
}

static void font_begin(TextFont *font) {
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, font->texture);
	glBegin(GL_QUADS);
}

static void font_end(void) {
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

static m4 state_transform(State *state) {
	return m4_ortho(0, state->width_gl, 0, 1, -1, +1);
}

static void sprites_begin(State *state) {
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, state->spritesheet.tex);
	glBegin(GL_QUADS);
}

static void sprites_end(void) {
	glEnd();
	glDisable(GL_TEXTURE_2D);
}

// set ghost to true when rendering a planned action, as opposed to an action we are planning right now 
static void render_action(State *state, Action *act, v4 color) {
	Sprite *sprites = state->sprites;
	Game *game = &state->game;
	uint kind = act->kind;

	color = v4_lerp(0.5f, color, V4(0.5f, 0.5f, 0.5f, 1.0f));

	switch (kind) {
	case ACT_NONE: break;
	case ACT_ATTACK:
	case ACT_MOVE: {
		HexPos hex1 = act->hex1, hex2 = act->hex2; 
		if (hex_in_bounds(game, hex1) && hex_in_bounds(game, hex2)) {
			Sprite *arrow = sprites + (kind == ACT_MOVE ? SPR_ARROW : SPR_ATTACK);
			float ax = arrow->x, ay = arrow->y, aw = arrow->w, ah = arrow->h;
			// arrow from hex1 to mouse hex 
			v2 from = hex_center_pos(game, hex1);
			v2 to = hex_center_pos(game, hex2);
			v2 v = v2_normalize(v2_sub(to, from));
			v2 right = V2(v.y, -v.x);
			v2 left  = V2(-v.y, v.x);
			v2 a,b,c,d; // corners of quad 
			float arrow_thickness = game->hex_radius * 0.3f;

			if (hex2.x == I16_MAX || hex2.y == I16_MAX) return;

			right = v2_scale(right, arrow_thickness);
			left  = v2_scale(left,  arrow_thickness);
			a = v2_add(from, left);
			b = v2_add(from, right);
			c = v2_add(to, right);
			d = v2_add(to, left);
			if (kind == ACT_ATTACK) {
				// tint attacks to make things clearer 
				color = v4_lerp(0.2f, color, V4(1,0.4f,0,1));
			}
			if (hexes_are_adjacent(hex1, hex2) && hex_can_have_pion(game, hex1) && hex_can_have_pion(game, hex2)) // this is not action_is_legal because that would be complicated
				v4_gl_color(color);
			else
				glColor4f(1,0,0,0.6f);
			glTexCoord2f(ax,    ay);    v2_gl_vertex(a);
			glTexCoord2f(ax,    ay+ah); v2_gl_vertex(b);
			glTexCoord2f(ax+aw, ay+ah); v2_gl_vertex(c);
			glTexCoord2f(ax+aw, ay);    v2_gl_vertex(d);
		}
	} break;
	}
}

static void render_hexes_start(State *state) {
	GLProcs *gl = &state->gl;
	HexProgram *hex_prog = &state->hex_program;
	m4 transform = state_transform(state);

	gl->UseProgram(hex_prog->program);
	gl->UniformMatrix4fv(hex_prog->transform, 1, GL_FALSE, transform.e);
	glBegin(GL_TRIANGLES);
}

static void render_hexes_end(State *state) {
	GLProcs *gl = &state->gl;
	glEnd();
	gl->UseProgram(0);
}

// make sure you are drawing triangles, and are using the hex program 
static void render_hex(State *state, v2 center, float hex_radius, v4 color) {
	HexProgram *prog = &state->hex_program;
	GLProcs *gl = &state->gl;

	v4 color_dark = v4_scale_xyz(color, 0.5f);
	int i;
	v2 points[6];
	static v2 const offsets[] = {
		{-0.5f, +HALF_SQRT3f},
		{+0.5f, +HALF_SQRT3f},
		{+1.0f, 0},
		{+0.5f, -HALF_SQRT3f},
		{-0.5f, -HALF_SQRT3f},
		{-1.0f, 0}
	};

	// calculate 6 hexagon vertices 
	for (i = 0; i < 6; ++i) {
		points[i] = v2_add(center, v2_scale(offsets[i], hex_radius));
	}
	v4_gl_color(color);
	// draw 6 hexagon triangles 
	for (i = 0; i < 6; ++i) {
		int j = (i+1)%6;
		v4_gl_color(color);
		gl->VertexAttrib1f(prog->bias, 0);
		v2_gl_vertex(center);

		v4_gl_color(i & 1 ? color_dark : color);
		gl->VertexAttrib1f(prog->bias, 1);
		v2_gl_vertex(points[i]);
		
		v4_gl_color(j & 1 ? color_dark : color);
		gl->VertexAttrib1f(prog->bias, 1);
		v2_gl_vertex(points[j]);
	}
}

// make sure you are rendering sprites 
static void render_pion(State *state, Game *game, Pion *pion, v4 color_mod) {
	float pion_size = game->hex_radius * 1.2f;
	v2 pion_dimensions = V2(pion_size, pion_size);
	v2 screen_pos = hex_center_pos(game, pion->pos);
	v4 color = v4_mul(player_colors[pion->player], color_mod);
	v4_gl_color(color);
	sprite_render_centered(state, SPR_PION, screen_pos, pion_dimensions);
}

// make sure you are rendering sprites
static void render_obj(State *state, ObjectKind obj, v2 center, float hex_radius, float alpha) {
	float obj_diameter = hex_radius * 1.2f;
	v2 obj_size = V2(obj_diameter, obj_diameter);

	switch (obj) {
	case OBJ_COUNT: case OBJ_NONE: break;
	case OBJ_BOX:
		glColor4f(0.5f, 0.3f, 0, alpha);
		sprite_render_centered(state, SPR_BOX, center, obj_size);
		break;
	
	}
}


// if NULL is passed for game, a grid with just TL_REGs and no pions,etc. will be rendered
static void render_hex_grid(State *state, Game *game, float hex_radius, v2 grid_base, float alpha, int width, int height, bool highlight_mouse_hex) {
	HexPos mouse_hex_pos = hex_at_pos_(state->mouse_pos, hex_radius, grid_base);
	uint mouse_mask = state->mouse_mask;
	int x, y;
	u8 *tile = game ? game->tiles : NULL, *obj = game ? game->objects : NULL;
	int grid_w = width, grid_h = height;

	render_hexes_start(state);

	for (y = 0; y < grid_h; ++y) {
		for (x = 0; x < grid_w; ++x) {
			v2 center = hex_center_pos2_(x, y, hex_radius, grid_base);
			static u8 const color_table[] = {
				0,1,
				2,0,
				1,2
			};
			v3 color;
			float c1 = 0.6f, c2 = 0.7f; // color values 
			uint tile_type = tile ? *tile : TL_REG;
			
			if (tile_type) {
				// pick hexagon color 
				switch (color_table[(y % 3) * 2 + (x % 2)]) {
				case 0:
					color = V3(c2, c1, c1);
					break;
				case 1:
					color = V3(c2, c2, c1);
					break;
				case 2:
					color = V3(c1, c1, c2);
					break;
				}

				if (highlight_mouse_hex && x == mouse_hex_pos.x && y == mouse_hex_pos.y) {
					v3 color_hl = V3(0,1,0);
					if (mouse_mask)
						color_hl = V3(0,0,1);
					color = v3_lerp(0.3f, color, color_hl);
				}
				
				render_hex(state, center, hex_radius, V4(color.x, color.y, color.z, alpha));
			}
			if (tile) ++tile;
		}
	}
	render_hexes_end(state);

	if (obj) {
		// render objects
		sprites_begin(state);
		for (y = 0; y < grid_h; ++y) {
			for (x = 0; x < grid_w; ++x) {
				if (*obj) {
					v2 center = hex_center_pos2_(x, y, hex_radius, grid_base);
					render_obj(state, *obj, center, hex_radius, alpha);
				}
				++obj;
			}
		}
		sprites_end();

	}


}

// render the hex grid and pions for the game 
// the alpha of everything is multiplied by alpha.
static void render_game_board(State *state, Game *game, float alpha, bool highlight_mouse_hex) {
	render_hex_grid(state, game, game->hex_radius, game->grid_base, alpha, game->width, game->height, highlight_mouse_hex);
	{
		Settings *settings = &state->settings;
		Pion *pion, *end;
		// draw pions 
		sprites_begin(state);
		for (pion = game->pions, end = pion + game->npions; pion != end; ++pion) {
			render_pion(state, game, pion, V4(1,1,1,alpha));
		}
		sprites_end();

		if (
		#if DEBUG
		game->number_pions || 
		#endif
		settings->label_pions
		) { // pion numbers 
			TextFont *font = &state->main_font;
			font_begin(font);
			for (pion = game->pions, end = pion + game->npions; pion != end; ++pion) {
				char text[32] = {0};
				v2 c = hex_center_pos(game, pion->pos);
				v4 color = v4_lerp(0.3f, player_colors[pion->player], V4(1, 1, 1, 1));
				if (settings->label_pions)
					sprintf(text, "%u", pion->player);
				#if DEBUG
				if (game->number_pions) {
					uint id = (uint)(pion - game->pions);
					sprintf(text + strlen(text), "%u", id);
				}
				#endif
				v4_gl_color(color);
				text_render_centered(font, text, c);
			}
			font_end();
		}

	}
}

