#ifdef DISABLE_NETWORKING
#include "sockets-none.c"
#else
#include <ifaddrs.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/random.h>
#include <fcntl.h>
#include <unistd.h>

#if DEBUG && 0
/*
	if this is set to 1, you can play a LAN game using the loopback interface (i.e. against yourself)
	this is for testing
*/
#define ALLOW_LOOPBACK_GAMES 1
#endif

static bool get_ip_address(IPv4 *ip) {
	struct ifaddrs *addrs, *a;
	getifaddrs(&addrs);
	for (a = addrs; a; a = a->ifa_next) {
		if (a->ifa_addr && a->ifa_addr->sa_family == AF_INET 
		#if !ALLOW_LOOPBACK_GAMES
			&& !(a->ifa_flags & IFF_LOOPBACK)
		#endif
			&& (a->ifa_flags & IFF_RUNNING)) {
			struct sockaddr_in *p = (struct sockaddr_in *)a->ifa_addr;
			u32 addr = p->sin_addr.s_addr;
		#if DEBUG
			printf("Found IP address %s\n",inet_ntoa(p->sin_addr));
		#endif
			ip->a = (u8)addr;
			ip->b = (u8)(addr >> 8);
			ip->c = (u8)(addr >> 16);
			ip->d = (u8)(addr >> 24);
			return true;
		}
	}
	freeifaddrs(addrs);
	return false;
}

struct Server {
	int socket, conn;
};

struct Client {
	int socket;
	bool connected;
	struct sockaddr_in addr;
};

static char comm_err[256];

static char *comm_geterr(void) {
	return comm_err;
}

static void comm_seterr(char const *fmt, ...) {
	va_list args;
	va_start(args, fmt);
	vsprintf(comm_err, fmt, args);
	va_end(args);
}

#if DEBUG
#define SC_DEBUG 1
#endif

static Server *server_create(ServerInfo *info) {
	Server *s = (Server *)calloc(1, sizeof *s);
	
	
	{ // set info->ip 
		IPv4 *ip = &info->ip;
		bool found_ip = get_ip_address(ip);
		if (!found_ip) {
			comm_seterr("Couldn't get IP address. Are you connected to the internet?");
			free(s);
			return NULL;
		}
	}
	{ // set info->password 
		info->password = (u16)rand();
	}

	if (s) {
		int sock = s->socket = socket(AF_INET, SOCK_STREAM, 0);
		s->conn = -1;
		if (sock != -1) {
			u8 port_lower_4bits;
			for (port_lower_4bits = 0; port_lower_4bits < 0x10; ++port_lower_4bits) {
				struct sockaddr_in addr = {0};
				u16 port_host = info->port = (u16)(((u16)PORT_UPPER_PART << 4) + port_lower_4bits); 
				{ // this lets us use the same port multiple times without exiting 
					int yes = 1;
					setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof yes);
				}
				fcntl(sock, F_SETFL, O_NONBLOCK);
				addr.sin_family = AF_INET;
				addr.sin_port = htons(port_host);
				addr.sin_addr.s_addr = htonl(INADDR_ANY);
				if (bind(sock, (struct sockaddr *)&addr, sizeof addr) != -1) {
					if (listen(sock, 10) != -1) {
						// all good 
						return s;
					} else {
						comm_seterr("Socket couldn't listen on port %u (%s)", port_host, strerror(errno));
						break;
					}
				} else if (errno == EADDRINUSE) {
					// just continue along; try the next port 
				} else {
					comm_seterr("Couldn't bind socket to port %u (%s)", port_host, strerror(errno));
					break;
				}
			}
			if (port_lower_4bits == 0x10) {
				// all 16 ports we tried were in use 
				comm_seterr("No ports are available.");
			}
			close(sock);
		} else {
			comm_seterr("Couldn't create socket.");
		}
		free(s);
	} else {
		comm_seterr("Out of memory.");
	}
	return NULL;
}

static bool server_check_for_client(Server *server, ClientInfo *info) {
	int conn = accept(server->socket, NULL, NULL);
	if (conn == -1) {
		// we can try again later 
		return false;
	}
	fcntl(conn, F_SETFL, O_NONBLOCK);
	server->conn = conn;
	info->unused = 0;
	return true;
}

static int deal_with_error(int err) {
	if (err == EAGAIN || err == EWOULDBLOCK || err == ENOBUFS) {
		return 0;
	} else {
		comm_seterr("Network error: %s", strerror(err));
		return -1;
	}
}

static long server_send(Server *server, void const *data, size_t n) {
	ssize_t nwritten = send(server->conn, data, n, MSG_NOSIGNAL);
#if SC_DEBUG
	if (nwritten > 0) {
		ssize_t i;
		printf("Server sent %ld bytes: ", (long)nwritten);
		for (i = 0; i < nwritten; ++i) {
			printf("%x ", ((u8 *)data)[i]);
		}
		printf("\n");
	}
#endif
	if (nwritten == -1) {
		return deal_with_error(errno);
	}
	return (long)nwritten;
}
static long server_recv(Server *server, void *data, size_t n) {
	ssize_t nread = read(server->conn, data, n);
#if SC_DEBUG
	if (nread > 0) {
		ssize_t i;
		printf("Server received %ld bytes: ", (long)nread);
		for (i = 0; i < nread; ++i) {
			printf("%x ", ((u8 *)data)[i]);
		}
		printf("\n");
	}
#endif
	if (nread == -1) {
		return deal_with_error(errno);
	}
	return (long)nread;
}

static void server_disconnect_from_client(Server *server) {
	close(server->conn);
	server->conn = -1;
}

static void server_destroy(Server *server) {
	if (server->conn) close(server->conn);
	if (server->socket) close(server->socket);
	free(server);
}

static Client *client_create(ServerInfo *server_info) {
	Client *c = (Client *)calloc(1, sizeof *c);
	int sock;
	if (c) {
		sock = c->socket = socket(AF_INET, SOCK_STREAM, 0);
		if (sock != -1) {
			struct sockaddr_in addr = {0};
			IPv4 ip = server_info->ip;
			u16 port_host = server_info->port;
			fcntl(sock, F_SETFL, O_NONBLOCK);
			addr.sin_family = AF_INET;
			addr.sin_port = htons(port_host);
			addr.sin_addr.s_addr = ((u32)ip.d << 24) | ((u32)ip.c << 16) | ((u32)ip.b << 8) | ip.a;
			c->addr = addr;
			if (connect(sock, (struct sockaddr *)&addr, sizeof addr) != -1) {
				c->connected = true;
				return c;
			} else if (errno == EAGAIN || errno == EINPROGRESS) {
				// not connected yet 
				return c;
			}
			comm_seterr("Couldn't connect to server (%s).", strerror(errno));
			close(sock);
		} else {
			comm_seterr("Couldn't create socket.");
		}
		free(c);
	} else {
		comm_seterr("Out of memory.");
	}
	return NULL;
}

static int client_is_connected(Client *client) {
	struct sockaddr_in addr = client->addr;
	if (client->connected) return 1;
	if (connect(client->socket, (struct sockaddr *)&addr, sizeof addr) != -1) {
		client->connected = true;
		return 1;
	} else if (errno == EAGAIN || errno == EINPROGRESS || errno == EALREADY) {
		return 0;
	} else {
		comm_seterr("Couldn't connect to server (%s).", strerror(errno));
		return -1;
	}
}

static long client_recv(Client *client, void *data, size_t n) {
	ssize_t nread = read(client->socket, data, n);
#if SC_DEBUG
	if (nread > 0) {
		ssize_t i;
		printf("Client received %ld bytes: ", (long)nread);
		for (i = 0; i < nread; ++i) {
			printf("%x ", ((u8 *)data)[i]);
		}
		printf("\n");
	}
#endif
	if (nread == -1) {
		return deal_with_error(errno);
	}
	return (long)nread;
}

static long client_send(Client *client, void const *data, size_t n) {
	ssize_t nwritten = send(client->socket, data, n, MSG_NOSIGNAL);
#if SC_DEBUG
	if (nwritten > 0) {
		ssize_t i;
		printf("Client sent %ld bytes: ", (long)nwritten);
		for (i = 0; i < nwritten; ++i) {
			printf("%x ", ((u8 *)data)[i]);
		}
		printf("\n");
	}
#endif
	if (nwritten == -1) {
		return deal_with_error(errno);
	}
	return (long)nwritten;
}

static void client_destroy(Client *client) {
	if (client->socket) close(client->socket);
	free(client);
}
#endif // DISABLE_NETWORKING 
