#define MENU_BASE_Y 0.8f

// returns which option was chosen by this event, or 0 if none was 
static int selmenu_event(SelectMenu *menu, State *state, Event *e) {
	float center_x = state->width_gl * 0.5f;
	float delta_y = 0.1f;
	v2 base_pos = V2(center_x, MENU_BASE_Y), pos = base_pos;
	v2 mouse_pos = state->mouse_pos;
	int chosen_option = -1, noptions = menu->noptions;
	TextFont *font = &state->big_font;
	char const *const * options = menu->options;
	switch (e->kind) {
	case EVENT_KEY_DOWN: {
		KeyEvent *k = &e->u.key;
		switch (k->key) {
		case KEY_UP:
			--menu->sel;
			if (menu->sel < 0) menu->sel += noptions;
			break;
		case KEY_DOWN:
			++menu->sel;
			if (menu->sel >= noptions) menu->sel -= noptions;
			break;
		case KEY_ENTER:
			chosen_option = menu->sel;
			break;
		}
	} break;
	case EVENT_MOUSE_DOWN: {
		MouseEvent *m = &e->u.mouse;
		if (m->button == MBUTTON_LEFT) {
			int i, sel = -1;
			for (i = 0; i < noptions; ++i) {
				char const *text = options[i];
				v2 size = get_text_size(font, text);
				if (centered_rect_contains_point(pos, V2(1.4f, size.y), mouse_pos)) {
					sel = i;
				}
				pos.y -= delta_y;
			}
			if (sel != -1) {
				chosen_option = sel;
			}
		}
	} break;
	default: break;
	}
	return chosen_option;
}

static void selmenu_render(SelectMenu *menu, State *state) {
	float center_x = state->width_gl * 0.5f;
	float delta_y = 0.1f;
	v2 base_pos = V2(center_x, MENU_BASE_Y), pos = base_pos;
	v2 mouse_pos = state->mouse_pos;
	TextFont *font = &state->big_font;
	font_begin(font);
	{
		int i;
		int sel = menu->sel, noptions = menu->noptions;
		char const* const* options = menu->options;

		for (i = 0; i < noptions; ++i) {
			char const *text = options[i];
			v2 size = get_text_size(font, text);
			if (centered_rect_contains_point(pos, V2(1.4f, size.y), mouse_pos)) {
				sel = i;
			}
			pos.y -= delta_y;
		}
		menu->sel = sel;
		pos = base_pos;
		for (i = 0; i < noptions; ++i) {
			char const *text = options[i];
			if (i == sel) {
				glColor3f(0.4f, 0.4f, 1);
			} else {
				glColor3f(1,1,1);
			}
			text_render_centered(font, text, pos);
			pos.y -= delta_y;
		}
	}
	font_end();
}

static TextFont *button_font(Button *button, State *state) {
	return button->font ? button->font : &state->main_font;
}

static v2 button_size(Button *button, State *state) {
	if (button->text) {
		return get_text_size(button_font(button, state), button->text);
	} else {
		return button->size;
	}
}

static v2 button_pos(Button *button, v2 button_size) {
	v2 bpos = button->pos;
	if (button->center) {
		bpos = v2_sub(bpos, v2_scale(button_size, 0.5f));
	}
	return bpos;
}

// font must be in use (see font_begin/end) 
static void button_render(Button *button, State *state) {
	{
		v2 mpos = state->mouse_pos;
		v2 bsize = button_size(button, state);
		v2 bpos = button_pos(button, bsize);
		bool has_color = button->color.w != 0;

		if (button->disabled) {
			glColor3f(0.3f, 0.3f, 0.3f);
		} else if (rect_contains_point_v2(bpos, bsize, mpos)) {
			if (has_color) {
				v4_gl_color(v4_scale(button->color, 0.7f));
			} else {
				glColor3f(0.4f, 0.4f, 1);
			}
		} else {
			if (has_color) {
				v4_gl_color(button->color);
			} else {
				gl_color_text();
			}
		}
	}
	if (button->text) {
		TextFont *font = button_font(button, state);
		if (button->center)
			text_render_centered(font, button->text, button->pos);
		else
			text_render(font, button->text, button->pos);
	} else {
		if (button->center)
			sprite_render_direction_centered(state, button->sprite, button->direction, button->pos, button->size);
		else
			sprite_render_direction(state, button->sprite, button->direction, button->pos, button->size);
	}
}

// returns true if button was pressed 
static bool button_event(Button *button, State *state, Event *e) {
	if (button->disabled) return false;
	if (e->kind == EVENT_MOUSE_DOWN) {
		MouseEvent *m = &e->u.mouse;
		if (m->button == MBUTTON_LEFT) {
			v2 mpos = pixels_to_gl_coords(state, m->x, m->y);
			v2 bsize = button_size(button, state);
			v2 bpos = button_pos(button, bsize);
			if (rect_contains_point_v2(bpos, bsize, mpos)) {
				return true;
			}
		}
	}
	return false;
}

static TextFont *cbox_font(Checkbox *checkbox, State *state) {
	return checkbox->font ? checkbox->font : &state->main_font;
}

#define CHECKBOX_SIZE 0.04f

static void checkbox_render(Checkbox *checkbox, State *state) {
	v2 mpos = state->mouse_pos;
	v2 cpos = checkbox->pos;
	v2 csize = V2(CHECKBOX_SIZE, CHECKBOX_SIZE);
	v3 color = V3(1,1,1);
	
	if (rect_contains_point_v2(cpos, csize, mpos)) {
		color = V3(0.4f, 0.4f, 1);
	}

	v3_gl_color(color);
	sprite_render(state, checkbox->checked ? SPR_CBOX_CHECKED : SPR_CBOX, cpos, csize);
}

static void checkbox_render_lbl(Checkbox *checkbox, State *state, char const *lbl) {
	TextFont *font = cbox_font(checkbox, state);
	v2 pos = checkbox->pos;
	v2 text_size = get_text_size(font, lbl);
	pos.x += CHECKBOX_SIZE + 0.01f;
	pos.y += (CHECKBOX_SIZE - text_size.y) * 0.5f;
	glColor3f(1,1,1);
	text_render(font, lbl, pos);
}

// returns true if checkbox->checked was changed 
static bool checkbox_event(Checkbox *checkbox, State *state, Event *e) {
	if (e->kind == EVENT_MOUSE_DOWN) {
		MouseEvent *m = &e->u.mouse;
		if (m->button == MBUTTON_LEFT) {
			v2 mpos = pixels_to_gl_coords(state, m->x, m->y);
			v2 cpos = checkbox->pos;
			v2 csize = V2(CHECKBOX_SIZE, CHECKBOX_SIZE);
			if (rect_contains_point_v2(cpos, csize, mpos)) {
				checkbox->checked = !checkbox->checked;
				return true;
			}
		}
	}
	return false;
}

static u16 text_input_max_len(TextInput *in) {
	if (in->max_len == 0) {
		return 32; // default max len 
	} else if (in->max_len > TEXT_INPUT_BUFSZ-1) { // check to make sure we don't have a potential overflow 
		assert(0);
		return TEXT_INPUT_BUFSZ-1;
	} else {
		return in->max_len;
	}
}

static TextFont *text_input_font(TextInput *in, State *state) {
	return in->font ? in->font : &state->main_font;
}

// rectangle of text input 
static Rect text_input_bbox(TextInput *in, State *state) {
	TextFont *font = text_input_font(in, state);
	float char_height = font_char_height(font);
	float x1 = in->pos.x, y1 = in->pos.y;
	float width = in->width == 0 ? 0.3f : in->width;
	float height = char_height;
	height += char_height * 0.4f;
	if (in->center) {
		return rect_centered(V2(x1,y1), V2(width,height));
	} else {
		return rect(V2(x1,y1), V2(width,height));
	}
}

static void text_input_render(TextInput *in, State *state) {
	Rect bbox = text_input_bbox(in, state);
	float bbx1,bby1,bbx2,bby2; // coordinates of bounding box 
	TextFont *font = text_input_font(in, state);
	float char_height = font->char_height_gl;
	char *text = in->text;
	float text_width = 0, text_x2;
	float cursor_width = 0.0025f;
	v4 color = in->color.w == 0 ? V4(0.8f, 0.8f, 0.8f, 1.0f) : in->color;

	rect_coords(bbox, &bbx1, &bby1, &bbx2, &bby2);
	
	text_width = get_text_size(font, text).x;

	if (bbx1 + text_width > bbx2 - cursor_width)
		text_x2 = bbx2 - cursor_width;
	else
		text_x2 = bbx1 + text_width;
	
	glBegin(GL_QUADS);
	{
		float mul = 0.5f;
		v4 box_color;
		if (in->focus)
			mul = 0.4f;
		else if (rect_contains_point(bbox, state->mouse_pos))
			mul = 0.8f;

		box_color = v4_scale(color, mul);
		box_color.w *= 0.9f;
		v4_gl_color(box_color);
		glVertex2f(bbx1, bby1);
		glVertex2f(bbx2, bby1);
		glVertex2f(bbx2, bby2);
		glVertex2f(bbx1, bby2);

		gl_color_text();
		if (in->focus) { // cursor 
			float x1 = text_x2, x2 = x1 + cursor_width;
			float y1 = bby1, y2 = bby2;
			glVertex2f(x1, y1);
			glVertex2f(x2, y1);
			glVertex2f(x2, y2);
			glVertex2f(x1, y2);
		}
	}
	glEnd();

	font_begin(font);
	{
		float x = text_x2 - text_width, y = bby1 + char_height * 0.15f;
		char const *str = text;
		v4 text_color = color;
		if (in->placeholder && !in->focus && text[0] == '\0') {
			// show placeholder text instead 
			str = in->placeholder;
			text_color.w = 0.3f;
		}
		v4_gl_color(text_color);
		text_render_(font, str, &x, &y, true, bbx1);
	}
	font_end();
}

// @TODO(eventually): arrow keys to move cursor
// returns true if user pressed enter 
static bool text_input_event(TextInput *in, State *state, Event *e) {
	char *text = in->text;
	switch (e->kind) {
	case EVENT_TEXT: {
		if (in->focus) {
			TextEvent *t = &e->u.text;
			if (str_satisfies(t->text, in->char_predicate ? in->char_predicate : isprint)) { // check that inputted text is valid
				str_cat(text, (size_t)text_input_max_len(in)+1, t->text); // add inputted text
			}
		}
	} break;
	case EVENT_KEY_DOWN:
		if (in->focus) {
			KeyEvent *k = &e->u.key;
			if (k->key == KEY_BACKSPACE) {
				if (*text) { // make sure text isn't empty
					if (state->control_key_down) {
						// delete word 
						int i;
						int stage = 0; // stage 0 = delete spaces after word, stage 1 = delete word, stage 2 = delete spaces before word, 3 = done 
						for (i = (int)strlen(text)-1; i >= 0 && stage <= 2; --i) {
							char c = text[i];
							text[i] = '\0';
							switch (stage) {
							case 0:
								if (!isspace(c))
									stage = 1;
								break;
							case 1:
								if (isspace(c))
									stage = 2;
								break;
							case 2:
								if (!isspace(c)) {
									text[i] = c;
									stage = 3;
								}
								break;
							}
						}
					} else {
						// delete char 
						text[strlen(text)-1] = '\0';
					}
				}
			} else if (k->key == KEY_ENTER) {
				in->focus = false;
				return true;
			}
		}
		break;
	case EVENT_MOUSE_DOWN: {
		MouseEvent *m = &e->u.mouse;
		v2 pos = pixels_to_gl_coords(state, m->x, m->y);
		Rect bbox = text_input_bbox(in, state);
		bool inside_bbox = rect_contains_point(bbox, pos);
		if (in->focus) {
			if (!inside_bbox)
				in->focus = false;
		} else {
			if (inside_bbox)
				in->focus = true;
		}
	} break;
	}
	return false;
}

static void text_input_clear(TextInput *input) {
	memset(input->text, 0, sizeof input->text);
}


// make sure that input contains an integer from min to max, settling on dfl if something's wrong.
static void text_input_validate_int(TextInput *input, i32 min, i32 max, i32 dfl) {
	char *text = input->text;
	// remove leading zeroes
	size_t leading_zeroes = strspn(text, "0");
	if (leading_zeroes) {
		if (leading_zeroes == strlen(text)) // text is entirely 0s
			strcpy(text, "0");
		else
			memmove(text, text + leading_zeroes, strlen(text) - leading_zeroes); // text has some leading zeroes
	}

	assert(min <= max);

	{
		bool is_a_num = false;
		i32 number = str_to_i32(text, &is_a_num);
		if (!is_a_num) {
			sprintf(text, "%ld", (long)dfl);
		} else {
			if (number < min)
				sprintf(text, "%ld", (long)min);
			else if (number > max)
				sprintf(text, "%ld", (long)max);
		}
	}
}
